@@ -3,7 +3,9 @@
 # system imports:
 import json
 import logging
+import socket
 from decimal import Decimal
+import ssl
 
 # thirdparty imports:
 import requests
@@ -45,6 +47,8 @@ def call_api(data, server_url=None, api_user=None, api_password=None):
     else:
         servers = settings.RIPPLE_API_DATA
 
+    timeouts = 0
+
     for server_config in servers:
         url = server_config.get('RIPPLE_API_URL', '')
         user = server_config.get('RIPPLE_API_USER', '')
@@ -53,7 +57,13 @@ def call_api(data, server_url=None, api_user=None, api_password=None):
         try:
             response = requests.post(
                 url, json.dumps(data), auth=auth, verify=False, timeout=5)
-        except requests.ConnectionError, e:
+        except TypeError:  # e.g. json encode error
+            raise
+        except (requests.exceptions.Timeout, ssl.SSLError):
+            timeouts += 1
+            error = e
+            continue
+        except Exception as e:
             error = e
             continue
 
@@ -72,6 +82,10 @@ def call_api(data, server_url=None, api_user=None, api_password=None):
                 result.get('error_message', 'no_message'),
             )
             continue
+
+    if timeouts == len(servers):
+        raise RippleApiError('Timeout', '', 'rippled timed out')
+
     raise error
 
 
@@ -215,17 +229,103 @@ def submit(tx_blob, server_url=None, api_user=None, api_password=None):
 
 
 def balance(
-        account, issuer, currency,
+        account, issuers, currency,
         server_url=None, api_user=None, api_password=None):
     results = call_api(
         {
             'method': 'account_lines',
-            'params': [{'account': account, 'peer': issuer}]
+            'params': [{'account': account}]
         },
         server_url=None, api_user=None, api_password=None
     )
     total = Decimal('0.0')
     for line in results['lines']:
-        if line['currency'] == currency and (line['account'] == issuer or issuer is None):
+        if line['currency'] == currency and (line['account'] in issuers or issuers is None):
             total += Decimal(line['balance'])
     return total
+
+
+def is_trust_set(trusts, peer, currency='', limit=0):
+    """
+    checks if 'trusts' trusts 'peer' with specified currency and limit
+
+
+    Params:
+        `trusts`:
+            ripple address, that trusts or not 'peer'
+        `peer`:
+            ripple_address, that is trusted by  by 'trusts'
+        `currency` (optional):
+            currency in which trust should be set
+        `limit` (optional):
+            minimal amount of trust
+
+    Returns boolean
+
+    """
+    trust_result = False
+
+    trust_lines = call_api(
+        {
+            'method': 'account_lines',
+            'params': [{
+                'account': trusts,
+                'peer': peer
+            }]
+        },
+    )
+
+    status = trust_lines['status'] == 'success'
+    trusts = trust_lines['lines']
+    if status and trusts and not currency:
+        trust_result = True
+
+    elif status and trusts and currency:
+        for trust in trusts:
+
+            if currency == trust['currency']:
+                trust_result = float(limit) <= float(trust['limit'])
+                break
+
+    return trust_result
+
+
+def book_offer(
+        taker_pays_curr, taker_pays_curr_issuer, taker_gets_curr, taker_gets_curr_issuer, taker_address='',
+        ledger='current', marker='', autobridge=True, server_url=None, api_user=None, api_password=None):
+    """
+    Gets currency exchange rates
+
+    Params:
+        'taker_pays':
+            Specified in the following forms 'XRP' or 'USD/rHb9CJAWyB4rj91VRWn96DkukG4bwdtyTh'
+            The currency and issuer the taker pays. Do not specify an issuing account if the currency is XRP.
+        'taker_gets':
+            Specified in the following forms 'XRP' or 'USD/rHb9CJAWyB4rj91VRWn96DkukG4bwdtyTh'
+            The currency and issuer the taker pays. Do not specify an issuing account if the currency is XRP.
+        'ledger' (optional):
+            "current" (default). "closed", "validated", ledger_index, or ledger.
+        'taker' (optional):
+            The address of the taker. This affects the funding of offers by owners as they may need to pay transfer fees.
+            For a neutral point of view specify ADDRESS_ONE (rrrrrrrrrrrrrrrrrrrrBZbvji).
+        'marker' (optional):
+            Specify the paging marker as JSON. Defaults to "".
+            Token indicating start of page, it is returned from a previous invocation.
+        'autobridge' (optional):
+            If present, specifies synthesize orders through XRP books. Defaults to true
+    """
+    taker_pays = 'XRP' if taker_pays_curr == 'XRP' else {"currency": taker_pays_curr, "issuer": taker_pays_curr_issuer}
+    taker_gets = 'XRP' if taker_gets_curr == 'XRP' else {"currency": taker_gets_curr, "issuer": taker_gets_curr_issuer}
+    data = {
+        "method": "book_offers",
+        "params": [{
+            "taker_pays": taker_pays,
+            "taker_gets": taker_gets,
+            "ledger": ledger,
+            "marker": marker,
+            "autobridge": autobridge}]
+        }
+    if taker_address:
+        data["params"][0]["taker"] = taker_address
+
+    return call_api(data, server_url, api_user, api_password)
